/*
 * PROJECT: matereal at http://mr.digitalmuseum.jp/
 * ----------------------------------------------------------------------------
 *
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is matereal.
 *
 * The Initial Developer of the Original Code is Jun KATO.
 * Portions created by the Initial Developer are
 * Copyright (C) 2009 Jun KATO. All Rights Reserved.
 *
 * Contributor(s): Jun KATO
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either of the GNU General Public License Version 2 or later (the "GPL"),
 * or the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 */
package jp.digitalmuseum.mr.core;

import java.io.File;

import java.io.IOException;
import java.io.OutputStream;
import java.util.HashMap;
import java.util.HashSet;
import java.util.Map;
import java.util.Set;
import java.util.Map.Entry;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;
import javax.xml.transform.Transformer;
import javax.xml.transform.TransformerConfigurationException;
import javax.xml.transform.TransformerException;
import javax.xml.transform.TransformerFactory;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamResult;

import org.w3c.dom.Document;
import org.w3c.dom.Element;
import org.w3c.dom.Node;
import org.xml.sax.SAXException;


/**
 * Catalog class
 *
 * @author Jun KATO
 */
public class Catalog {
	private Map<Entity, Node> entityNodeMap;
	private DocumentBuilder builder;
	private Transformer transformer;
	private Document doc;
	private Element root;

	/**
	 * Create a new catalog.
	 */
	public Catalog() {
		entityNodeMap = new HashMap<Entity, Node>();

		final DocumentBuilderFactory dbfactory = DocumentBuilderFactory.newInstance();
		try {
			builder = dbfactory.newDocumentBuilder();
		} catch (ParserConfigurationException e) {
			throw new IllegalStateException("Catalog couldn't build a DocumentBuilder.");
		}
		final TransformerFactory tfactory = TransformerFactory.newInstance();
		try {
			transformer = tfactory.newTransformer();
		} catch (TransformerConfigurationException e) {
			throw new IllegalStateException("Catalog couldn't build a Transformer.");
		}

		doc = builder.newDocument();
		root = doc.createElement("matereal");
		doc.appendChild(root);
	}

	/**
	 * Load catalog information from the specified URI.
	 *
	 * @param uri
	 * @throws SAXException
	 * @throws IOException
	 */
	public void load(String uri) throws SAXException, IOException {
		doc = builder.parse(uri);
		parse();
	}

	/**
	 * Load catalog information from the specified local file.
	 *
	 * @param file
	 * @throws SAXException
	 * @throws IOException
	 */
	public void loadFromLocal(File file) throws SAXException, IOException {
		doc = builder.parse(file);
		parse();
	}

	/**
	 * Load catalog information from a local file whose name specified.
	 * @param fileName
	 * @throws SAXException
	 * @throws IOException
	 */
	public void loadFromLocal(String fileName) throws SAXException, IOException {
		loadFromLocal(new File(fileName));
	}

	private void parse() {
		//
		//
	}

	/**
	 * Export this catalog as a stream of string.
	 *
	 * @param outStream
	 */
	public void export(OutputStream outStream) {
		if (doc == null) {
			throw new IllegalStateException(
					"Catalog document is null.");
		}
        try {
			transformer.transform(new DOMSource(doc), new StreamResult(outStream));
		} catch (TransformerException e) {
			e.printStackTrace();
		}
	}

	/**
	 * Import catalog information from running matereal.
	 *
	 * @return Catalog instance.
	 */
	public static Catalog build() {
		final Matereal matereal = Matereal.getInstance();
		final Catalog catalog = new Catalog();
		catalog.doc = catalog.builder.newDocument();

		// Append Entity related nodes.
		appendEntities(catalog, matereal);

		// Append Service related nodes.
		final Set<ServiceGroup> serviceGroups =
				new HashSet<ServiceGroup>(matereal.getServiceGroups());
		serviceGroups.add(matereal);
		for (ServiceGroup serviceGroup : serviceGroups) {

			// Set up a parent node.
			Element currentRoot;
			if (serviceGroup.equals(matereal)) {
				currentRoot = catalog.root;
			} else {
				currentRoot = catalog.doc.createElement("group");
				currentRoot.setAttribute("name", serviceGroup.getName());
				catalog.root.appendChild(currentRoot);
			}
			currentRoot.setAttribute("interval", String.valueOf(serviceGroup.getInterval()));

			// Append Service related nodes.
			for (Service service : serviceGroup.getServices()) {

				// matereal>service or matereal>group>service
				final Element serviceNode = catalog.doc.createElement("service");

				// TODO
				serviceNode.setAttribute("name", service.getName());
				currentRoot.appendChild(serviceNode);

				// matereal>entity>(Entity nodes)>(Service option nodes)
				/*
				final ServiceSettingHandler handler = service.getDialogHandler();
				for (Entry<Entity, Node> set : catalog.entityNodeMap.entrySet()) {
					final Node entityNode =
							handler.exportEntityRelatedNode(set.getKey());
					if (entityNode != null) {
						set.getValue().appendChild(node);
					}
				}
				*/
			}
		}
		return catalog;
	}

	private static void appendEntities(Catalog catalog, EntityHolder holder) {
		for (Entity e : holder.getEntities()) {
			final Node node = buildEntityNode(catalog.doc, e);
			catalog.root.appendChild(node);
			catalog.entityNodeMap.put(e, node);
		}
	}

	private static Node buildEntityNode(Document doc, Entity e) {
		Element entityNode = doc.createElement("entity");
		entityNode.setAttribute("class", e.getClass().getCanonicalName());
		entityNode.setAttribute("name", e.getName());
		return entityNode;
	}

}
