/*
 * PROJECT: matereal at http://mr.digitalmuseum.jp/
 * ----------------------------------------------------------------------------
 *
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is matereal.
 *
 * The Initial Developer of the Original Code is Jun KATO.
 * Portions created by the Initial Developer are
 * Copyright (C) 2009 Jun KATO. All Rights Reserved.
 *
 * Contributor(s): Jun KATO
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either of the GNU General Public License Version 2 or later (the "GPL"),
 * or the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 */
package jp.digitalmuseum.mr.core;

import java.util.Collection;

import java.util.HashMap;
import java.util.HashSet;
import java.util.Iterator;
import java.util.Set;


/**
 * Utility class providing core-related functions.
 *
 * @author Jun KATO
 */
public final class CoreUtils {

	/**
	 * @param object
	 * @param castingClassObject
	 * @return Returns if the class of the specified object can be casted to a class expressed by the class object.
	 */
	public static boolean objectCastableTo(Object object, Class<?> castingClassObject) {
		return classCastableTo(object.getClass(), castingClassObject);
	}

	/**
	 * @param classObject
	 * @param castingClassObject
	 * @return Returns if the specified class can be casted to a class expressed by the class object.
	 */
	public static boolean classCastableTo(Class<?> classObject, Class<?> castingClassObject) {
		return castingClassObject.isAssignableFrom(classObject);
	}

	/**
	 * @param requirements
	 * @param supply
	 * @return Returns if the requirements can be satisfied with the supply by means of class types.
	 */
	public static boolean requirementsSuppliable(
			Set<Class<? extends Resource>> requirements,
			Set<Class<? extends Resource>> supply) {
		final Set<Class<? extends Resource>> reqs
				= new HashSet<Class<? extends Resource>>(requirements);
		for (Class<? extends Resource> cls : supply) {
			for (Iterator<Class<? extends Resource>> it = reqs.iterator(); it.hasNext();) {
				final Class<? extends Resource> req = it.next();
				if (classCastableTo(cls, req)) {
					it.remove();
				}
			}
		}
		return reqs.isEmpty();
	}

	/**
	 * @param resourceTypes
	 * @param resources
	 * @return Returns a resource map for a robot.
	 * @see Robot#requestResources(Set)
	 */
	public static HashMap<Class<? extends Resource>, Resource>
			constructResourceMap(Set<Class<? extends Resource>> resourceTypes, Collection<Resource> resources) {
		final HashMap<Class<? extends Resource>, Resource> resourceMap = new HashMap<Class<? extends Resource>, Resource>();
		final HashSet<Class<? extends Resource>> copyOfResourceTypes = new HashSet<Class<? extends Resource>>(resourceTypes);
		for (Resource resource : resources) {
			for (Iterator<Class<? extends Resource>> it = copyOfResourceTypes.iterator(); it.hasNext();) {
				final Class<? extends Resource> i = it.next();
				if (CoreUtils.classCastableTo(resource.getClass(), i)) {
					it.remove();
					resourceMap.put(i, resource);
				}
			}
		}
		return copyOfResourceTypes.isEmpty() ? resourceMap : null;
	}

	public static Collection<Class<? extends Resource>> getClassesOfResources(Collection<? extends Resource> resources) {
		final HashSet<Class<? extends Resource>> classes = new HashSet<Class<? extends Resource>>();
		for (Resource resource : resources) {
			classes.add(resource.getClass());
		}
		return classes;
	}
}
