/*
 * PROJECT: matereal at http://mr.digitalmuseum.jp/
 * ----------------------------------------------------------------------------
 *
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is matereal.
 *
 * The Initial Developer of the Original Code is Jun KATO.
 * Portions created by the Initial Developer are
 * Copyright (C) 2009 Jun KATO. All Rights Reserved.
 *
 * Contributor(s): Jun KATO
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either of the GNU General Public License Version 2 or later (the "GPL"),
 * or the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 */
package jp.digitalmuseum.mr.core;

import java.io.IOException;

import java.io.PrintStream;
import java.util.HashSet;
import java.util.Set;
import java.util.concurrent.Executors;
import java.util.concurrent.ScheduledExecutorService;

import javax.swing.UIManager;

import org.xml.sax.SAXException;

/**
 * Main class of matereal. Singleton.
 *
 * @author Jun KATO
 */
public final class Matereal extends ServiceGroupAbstractImpl implements EntityHolder {
	final public static String LIBRARY_NAME = "Matereal";
	final private static String THREAD_NAME = "Matereal Thread";
	public static int NUM_THREADS = 5;
	private static Matereal instance;
	private ScheduledExecutorService executor;
	private Catalog catalog;
	private Set<PhysicalEntity> entities;
	private Set<EntityHolder> holders;
	private Set<ServiceGroupImpl> groups;
	private PrintStream out;
	private PrintStream err;
	private boolean isDisposing = false;

// Singleton and application related methods.

	/**
	 * Private constructor
	 */
	private Matereal() {
		super();
		System.out.println("--Starting matereal.");
		executor = Executors.newScheduledThreadPool(NUM_THREADS);
		entities = new HashSet<PhysicalEntity>();
		holders = new HashSet<EntityHolder>();
		groups = new HashSet<ServiceGroupImpl>();
		out = System.out;
		err = System.err;
		setInterval(DEFAULT_PERIOD);
		setLookAndFeel();
	}

	/**
	 * Get the singleton instance of Matereal.
	 * @return Returns the singleton instance of this class.
	 */
	public synchronized static Matereal getInstance() {
		if (instance == null) {
			instance = new Matereal();
		}
		return instance;
	}

	public String getName() {
		return THREAD_NAME;
	}

	public boolean isDisposing() {
		return isDisposing;
	}

	/**
	 * Stop all services and prepare for shutting down Matereal.
	 */
	@Override
	public synchronized void dispose() {
		if (isDisposing) {
			return;
		}
		isDisposing = true;
		System.out.println("--Shutting down matereal.");
		synchronized (groups) {
			for (ServiceGroupImpl group : groups) {
				group.dispose();
			}
			groups.clear();
		}
		synchronized (holders) {
			for (EntityHolder holder : holders) {
				holder.dispose();
			}
			holders.clear();
		}
		synchronized (entities) {
			for (PhysicalEntity physicalEntity : entities) {
				physicalEntity.dispose();
			}
			entities.clear();
		}
		super.dispose();
		executor.shutdown();
		System.out.println("--THANK YOU. GOOD DAY!");
		isDisposing = false;
	}

// Core functions.

	/**
	 * Get a set of entity holders
	 * @return
	 */
	public Set<EntityHolder> getEntityHolders() {
		synchronized (holders) {
			return new HashSet<EntityHolder>(holders);
		}
	}

	/**
	 * Get a set of entities managed by matereal.
	 */
	public Set<Entity> getEntities() {
		synchronized (entities) {
			return new HashSet<Entity>(entities);
		}
	}

	/**
	 * Get running service groups.
	 * @return Returns current service groups.
	 */
	public Set<ServiceGroupImpl> getServiceGroups() {
		synchronized (groups) {
			return new HashSet<ServiceGroupImpl>(groups);
		}
	}

	/**
	 * Look for services with specific class/interface
	 * in all service groups (including matereal thread).
	 *
	 * @param <T> Class to look for.
	 * @param classObject Class object of the class <T>.
	 * @return
	 * @see ServiceGroupImpl
	 * @see #lookForServices(Class, ServiceGroup)
	 */
	public <T> Set<T> lookForAllServices(Class<T> classObject) {

		// Look for Matereal thread.
		Set<T> services = lookForServices(classObject);

		// Look for other service groups.
		synchronized (groups) {
			for (ServiceGroupImpl group : groups) {
				services.addAll(
						lookForServices(classObject, group));
			}
		}

		return services;
	}

	/**
	 * Look for services with specific class/interface
	 * in the matereal thread.
	 *
	 * @param <T> Class to look for.
	 * @param classObject Class object of the class <T>.
	 * @return
	 * @see ServiceGroupImpl
	 * @see #lookForAllServices(Class)
	 */
	public <T> Set<T> lookForServices(Class<T> classObject) {

		// Look for matereal thread.
		return lookForServices(classObject, this);
	}

	/**
	 * Look for services with specific class/interface in a service group.
	 *
	 * @param <T> Class to look for.
	 * @param classObject Class object of the class <T>.
	 * @param serviceGroup ServiceGroup to look in.
	 * @return Returns services satisfying given conditions.
	 * @see #lookForAllServices(Class)
	 */
	public <T> Set<T> lookForServices(Class<T> classObject, ServiceGroup serviceGroup) {
		Set<T> services = new HashSet<T>();
		synchronized (serviceGroup) {
			for (Service service : serviceGroup) {
				if (CoreUtils.objectCastableTo(service, classObject)) {
					services.add(classObject.cast(service));
				}
			}
		}
		return services;
	}

	ScheduledExecutorService getThreadPoolExecutor() {
		return executor;
	}

	/**
	 * Called by ServiceGroup constructor.
	 * @param group
	 * @see ServiceGroupImpl
	 */
	void registerServiceGroup(ServiceGroupImpl group) {
		synchronized (groups) {
			groups.add(group);
		}
	}

	/**
	 * Called by ServiceGroupAbstractImpl.dispose().
	 * @param group
	 * @return Returns unregistered ServiceGroup if succeeded.
	 * @see ServiceGroupImpl#dispose()
	 */
	ServiceGroupImpl unregisterServiceGroup(ServiceGroupImpl group) {
		synchronized (groups) {
			return groups.remove(group) ? group : null;
		}
	}

	/**
	 * Called by PhysicalEntityAbstractImpl constructor.
	 * @param entity
	 * @see PhysicalEntityAbstractImpl
	 */
	void registerEntity(PhysicalEntity entity) {
		synchronized (entities) {
			entities.add(entity);
		}
	}

	/**
	 * Called by PhysicalEntityAbstractImpl.dispose().
	 * @param entity
	 * @return Returns unregistered entity if succeeded.
	 */
	Entity unregisterEntity(PhysicalEntity entity) {
		synchronized (entities) {
			return entities.remove(entity) ? entity : null;
		}
	}

	/**
	 * Called by EntityHolder constructor.
	 * @param holder
	 * @see ProxyRobotHolder
	 */
	void registerEntityHolder(EntityHolder holder) {
		synchronized (holders) {
			holders.add(holder);
		}
	}

	/**
	 * Called by EntityHolderAbstractImpl.dispose().
	 * @param entityHolder
	 * @return Returns unregistered entity holder if succeeded.
	 */
	EntityHolder unregisterEntityHolder(EntityHolder entityHolder) {
		synchronized (entities) {
			return holders.remove(entityHolder) ? entityHolder : null;
		}
	}

	/**
	 * Load catalog from a file.
	 * @param fileName
	 */
	public synchronized boolean loadCatalog(String fileName) {
		catalog = new Catalog();
		try {
			catalog.loadFromLocal(fileName);
		} catch (SAXException e) {
			err.print("XML parse error during loadCatalog(");
			err.print(fileName);
			err.println(").");
			return false;
		} catch (IOException e) {
			err.print("I/O error during loadCatalog(");
			err.print(fileName);
			err.println(").");
			return false;
		}
		out.print("Successfully loaded a catalog(");
		err.print(fileName);
		err.println(").");
		return true;
	}

	/**
	 * Set look and feel of swing components.
	 */
	private void setLookAndFeel() {
		/*
		final LookAndFeelInfo[] lafs =
			UIManager.getInstalledLookAndFeels();
		String className = null;
		for (LookAndFeelInfo lafInfo : lafs) {
			final String name = lafInfo.getName();

			// Windows Look and Feel
			if (name.equals("Windows")) {
				className = lafInfo.getClassName();
				break;
			}

			// Windows Classic? Look and Feel
			if (name.indexOf("Windows") >= 0
					&& className == null) {
				className = lafInfo.getClassName();
				continue;
			}

			// Mac OSX Look and Feel
			if (name.indexOf("Mac") >= 0) {
				className = lafInfo.getClassName();
				continue;
			}
		}
		*/
		try {
			// UIManager.setLookAndFeel(className);
			UIManager.setLookAndFeel(
					UIManager.getSystemLookAndFeelClassName());
		} catch (Exception e) {
			// Do nothing.
		}
	}

	@Override
	public String toString() {
		return LIBRARY_NAME;
	}

}
