/*
 * PROJECT: matereal at http://mr.digitalmuseum.jp/
 * ----------------------------------------------------------------------------
 *
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is matereal.
 *
 * The Initial Developer of the Original Code is Jun KATO.
 * Portions created by the Initial Developer are
 * Copyright (C) 2009 Jun KATO. All Rights Reserved.
 *
 * Contributor(s): Jun KATO
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either of the GNU General Public License Version 2 or later (the "GPL"),
 * or the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 */
package jp.digitalmuseum.mr.core;

import java.awt.event.ActionListener;
import java.util.Iterator;
import java.util.List;
import java.util.concurrent.ScheduledExecutorService;
import java.util.concurrent.ScheduledFuture;
import java.util.concurrent.TimeUnit;

import javax.swing.Timer;

import jp.digitalmuseum.utils.VariableLengthArray;

/**
 * Abstract implementation of ServiceGroup.
 *
 * @author Jun KATO
 * @see ServiceGroup
 */
public abstract class ServiceGroupAbstractImpl implements ServiceGroup {
	final public static long DEFAULT_PERIOD = 33;
	private ScheduledFuture<?> timer;
	private Timer swingTimer;
	private long interval;
	private VariableLengthArray<Service> services;

	public ServiceGroupAbstractImpl() {
		services = new VariableLengthArray<Service>();
	}

	final public synchronized void setInterval(long interval) {
		this.interval = interval;
		if (timer != null) {
			timer.cancel(false);
		}
		timer = getThreadPoolExecutor().scheduleAtFixedRate(
				new ServiceTimerTask(),
				interval,
				interval,
				TimeUnit.MILLISECONDS);

		// Update Swing Timer
		if (swingTimer != null) {
			swingTimer.setDelay((int) interval);
		}
	}

	abstract ScheduledExecutorService getThreadPoolExecutor();

	final public synchronized long getInterval() {
		return interval;
	}

	final public synchronized void registerService(Service service) {
		services.push(service);
	}

	final public synchronized Service unregisterService(Service service) {
		return services.remove(service) ? service : null;
	}

	final public synchronized void mergeServiceGroup(ServiceGroupImpl serviceGroup) {
		final List<Service> services = serviceGroup.getServices();
		for (Service service : services) {
			serviceGroup.unregisterService(service);
			registerService(service);
		}
		serviceGroup.dispose();
	}

	final public synchronized List<Service> clearServices() {
		final List<Service> list = services.asList();
		for (Service service : services) {
			service.stop();
		}
		return list;
	}

	final public synchronized List<Service> getServices() {
		return services.asList();
	}

	final public void addActionListener(ActionListener listener) {

		// Instantiate a new Timer if needed, and add an ActionListener to the timer.
		if (swingTimer == null) {
			swingTimer = new javax.swing.Timer((int) getInterval(), listener);
		} else {
			swingTimer.addActionListener(listener);
		}

		// Start the timer if it is stopped.
		if (!swingTimer.isRunning()) {
			swingTimer.start();
		}
	}

	final public void removeActionListener(ActionListener listener) {
		if (swingTimer == null) {
			return;
		}
		swingTimer.removeActionListener(listener);

		// Stop the timer if no listeners are left.
		if (swingTimer.getActionListeners().length == 0) {
			swingTimer.stop();
		}
	}

	/**
	 * Stop and dispose this service group.
	 * Services registered to this group will be disposed, too.
	 *
	 * @see Matereal#dispose()
	 * @see ServiceGroupImpl#dispose()
	 */
	public synchronized void dispose() {
		timer.cancel(false);
		timer = null;
		if (swingTimer != null) {
			swingTimer.stop();
			swingTimer = null;
		}
		clearServices();
	}

	final public synchronized Iterator<Service> iterator() {
		return services.iterator();
	}

	/**
	 * Each service group is executed periodically as a TimerTask.
	 *
	 * @author Jun KATO
	 */
	final private class ServiceTimerTask implements Runnable {

		public void run() {
			synchronized (ServiceGroupAbstractImpl.this) {

				// Run each service,
				for (Service service : services) {

					// if it is not paused
					if (!service.isPaused()){
						service.run();
					}
				}
			}
		}

	}
}
