/*
 * PROJECT: matereal at http://mr.digitalmuseum.jp/
 * ----------------------------------------------------------------------------
 *
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is matereal.
 *
 * The Initial Developer of the Original Code is Jun KATO.
 * Portions created by the Initial Developer are
 * Copyright (C) 2009 Jun KATO. All Rights Reserved.
 *
 * Contributor(s): Jun KATO
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either of the GNU General Public License Version 2 or later (the "GPL"),
 * or the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 */
package jp.digitalmuseum.mr.entity;

import java.awt.Shape;
import java.awt.geom.Ellipse2D;
import java.awt.image.BufferedImage;
import java.io.IOException;
import java.net.MalformedURLException;
import java.net.URL;

import javax.imageio.ImageIO;

import jp.digitalmuseum.connector.Connector;
import jp.digitalmuseum.mr.core.Resource;
import jp.digitalmuseum.mr.core.ResourceAbstractImpl;
import jp.digitalmuseum.mr.core.RobotAbstractImpl;
import jp.digitalmuseum.mr.core.Service;
import jp.digitalmuseum.mr.core.ServiceAbstractImpl;
import jp.digitalmuseum.mr.core.ServiceGroup;
import jp.digitalmuseum.mr.core.ServiceGroupImpl;
import jp.digitalmuseum.mr.resource.DifferentialWheels;
import jp.digitalmuseum.mr.resource.HeadmountedCamera;

/**
 * NetTansor/NetTansor Web
 *
 * @author Jun KATO
 * @see NetTansorDriver
 */
public class NetTansor extends RobotAbstractImpl {
	final public double RADIUS = 8;
	NetTansorDriver driver;
	Resource[] resources;
	Shape shape;

	public NetTansor(String name) {
		super();
		initialize(name);
	}

	public NetTansor(String name, String connectionString) {
		super(connectionString);
		initialize(name);
	}

	public NetTansor(String name, Connector connector) {
		super(connector);
		initialize(name);
	}

	@Override
	public void dispose() {
		driver.stopWheels();
		super.dispose();
	}

	private void initialize(String name) {
		setName(name);
		setTypeName("NetTansor");
		driver = new NetTansorDriver(this);
		resources = new Resource[] { driver, new NetTansorHeadmountedCamera(this) };
		shape = new Ellipse2D.Double(-RADIUS, -RADIUS, RADIUS*2, RADIUS*2);
	}

	@Override
	protected Resource[] getResources() {
		return resources.clone();
	}

	public Shape getShape() {
		return shape;
	}

	/**
	 * Differential wheels of NetTansor.
	 *
	 * @author Jun KATO
	 * @see NetTansor
	 */
	public static class NetTansorDriver extends ResourceAbstractImpl implements DifferentialWheels {
		private static final int MAXIMUM_DISTANCE = 250;
		private static final int MAXIMUM_ANGLE = 360;
		private static final int MAXIMUM_POWER = 31;
		private URL url;
		private int distance;
		private int power;
		private int angle;

		public NetTansorDriver(NetTansor netTansor) {
			super(netTansor);
			initialize();
		}

		public NetTansorDriver(Connector connector) {
			super(connector);
			initialize();
		}

		private void initialize() {
			distance = MAXIMUM_DISTANCE;
			power = 15;
			angle = MAXIMUM_ANGLE;
			updateURL();
		}

		public void goBackward() {
			write(String.format("bk,%d,%d,0,0,0", power, distance));
		}

		public void goForward() {
			write(String.format("fd,%d,%d,0,0,0", power, distance));
		}

		public double getSpeed() {
			return 0;
		}

		public void setSpeed(double mmsec) {
			if(MAXIMUM_POWER < mmsec) {
				mmsec = MAXIMUM_POWER;
			}
			if(mmsec < 0) {
				mmsec = 0;
			}
			power = (int) mmsec;
		}

		public void spin(SPIN spin) {
			switch(spin) {
			case LEFT:
				write(String.format("lt,%d,%d,0,0,0", power, angle));
				break;
			case RIGHT:
				write(String.format("rt,%d,%d,0,0,0", power, angle));
				break;
			}
		}

		public void rotate(double radius, SPIN direction) {
			// TODO implement "rotate" for NetTansor
			/*
			write(String.format("mm,%d,%d",
				power,
				(int) (power *(1.0 - (double)p/100))));
			*/
			throw new UnsupportedOperationException();
		}

		public void stopWheels() {
			write("st");
		}

		private void write(String message) {
			getConnector().write(message);
			getConnector().write(0x0d);
		}

		public BufferedImage getImage() {
			try
			{
				return ImageIO.read(url);
			}
			catch(IOException e)
			{
				// e.printStackTrace();
				return null;
			}
		}

		private void updateURL() {
			try {
				url = new URL(getConnector().getConnectionString()+"/goform/video/");
			} catch (MalformedURLException e) {
				// e.printStackTrace();
				url = null;
			}
		}

		@Override
		protected void setConnector(Connector connector) {
			super.setConnector(connector);
			if (getConnector() != null) {
				updateURL();
			}
		}

		@Override
		protected void dispose() {
			stopWheels();
		}
	}

	public static class NetTansorHeadmountedCamera extends ResourceAbstractImpl implements HeadmountedCamera {
		final private ServiceGroup captureServiceGroup;
		final private Service captureService;
		private BufferedImage image;

		public NetTansorHeadmountedCamera(final NetTansor netTansor) {
			super(netTansor);
			captureServiceGroup = new ServiceGroupImpl();
			captureService = new ServiceAbstractImpl() {
				public void run() {
					synchronized (NetTansorHeadmountedCamera.this) {
						image = netTansor.driver.getImage();
					}
				}
			};
		}

		public synchronized BufferedImage getImage() {
			return image;
		}

		@Override
		protected void dispose() {
			captureService.stop();
		}

		public void startCapture() {
			captureService.start(captureServiceGroup);
		}

		public void endCapture() {
			captureService.stop();
		}
	}
}
