/*
 * PROJECT: matereal at http://mr.digitalmuseum.jp/
 * ----------------------------------------------------------------------------
 *
 * Version: MPL 1.1/GPL 2.0/LGPL 2.1
 *
 * The contents of this file are subject to the Mozilla Public License Version
 * 1.1 (the "License"); you may not use this file except in compliance with
 * the License. You may obtain a copy of the License at
 * http://www.mozilla.org/MPL/
 *
 * Software distributed under the License is distributed on an "AS IS" basis,
 * WITHOUT WARRANTY OF ANY KIND, either express or implied. See the License
 * for the specific language governing rights and limitations under the
 * License.
 *
 * The Original Code is matereal.
 *
 * The Initial Developer of the Original Code is Jun KATO.
 * Portions created by the Initial Developer are
 * Copyright (C) 2009 Jun KATO. All Rights Reserved.
 *
 * Contributor(s): Jun KATO
 *
 * Alternatively, the contents of this file may be used under the terms of
 * either of the GNU General Public License Version 2 or later (the "GPL"),
 * or the GNU Lesser General Public License Version 2.1 or later (the "LGPL"),
 * in which case the provisions of the GPL or the LGPL are applicable instead
 * of those above. If you wish to allow use of your version of this file only
 * under the terms of either the GPL or the LGPL, and not to allow others to
 * use your version of this file under the terms of the MPL, indicate your
 * decision by deleting the provisions above and replace them with the notice
 * and other provisions required by the GPL or the LGPL. If you do not delete
 * the provisions above, a recipient may use your version of this file under
 * the terms of any one of the MPL, the GPL or the LGPL.
 */
package jp.digitalmuseum.mr.entity;

import java.awt.Shape;
import java.awt.geom.RoundRectangle2D;


import jp.digitalmuseum.connector.Connector;
import jp.digitalmuseum.mr.core.Resource;
import jp.digitalmuseum.mr.core.ResourceAbstractImpl;
import jp.digitalmuseum.mr.core.RobotAbstractImpl;
import jp.digitalmuseum.mr.resource.Wheels;

/**
 * Noopy, a prototype robot.
 *
 * @author Jun KATO
 * @see NoopyWheels
 */
public class Noopy extends RobotAbstractImpl {
	final public static double WIDTH = 12;
	final public static double HEIGHT = 16;
	private Wheels w;
	private Resource[] resources;
	private Shape shape;

	public Noopy(String name) {
		super();
		initialize(name);
	}

	public Noopy(String name, String connectionString) {
		super(connectionString);
		initialize(name);
	}

	public Noopy(String name, Connector connector) {
		super(connector);
		initialize(name);
	}

	@Override
	public void dispose() {
		w.stopWheels();
		super.dispose();
	}


	private void initialize(String name) {
		setName(name);
		setTypeName("Noopy");
		w = new NoopyWheels(this);
		resources = new Resource[] { w };
		shape = new RoundRectangle2D.Double(
				-WIDTH/2 , -HEIGHT/2,
				WIDTH/2, HEIGHT/2,
				3, 3);
	}

	@Override
	protected Resource[] getResources() {
		return resources.clone();
	}

	public Shape getShape() {
		return shape;
	}

	/**
	 * Wheels of Noopy.
	 *
	 * @author Jun KATO
	 * @see Noopy
	 */
	public static class NoopyWheels extends ResourceAbstractImpl implements Wheels {

		private enum STATUS {
			GO_BACK, GO_FORWARD, SPIN_LEFT, SPIN_RIGHT, STOP
		}
		private enum PEN {
			UP, DOWN
		}
		private STATUS status;
		private PEN pen;

		public NoopyWheels(Noopy noopy) {
			super(noopy);
			initialize();
		}

		public NoopyWheels(Connector connector) {
			super(connector);
			initialize();
		}

		private void initialize() {
			status = STATUS.STOP;
			pen = PEN.UP;
		}

		@Override
		protected void dispose() {
			endPen();
			stopWheels();
		}

		public void goBackward() {
			if (!status.equals(STATUS.GO_BACK)) {
				write("i");
				status = STATUS.GO_BACK;
			}
		}

		public void goForward() {
			if (!status.equals(STATUS.GO_FORWARD)) {
				write("d");
				status = STATUS.GO_FORWARD;
			}
		}

		public void spin(SPIN direction) {
			if (direction.equals(SPIN.LEFT)) {
				if (!status.equals(STATUS.SPIN_LEFT)) {
					write("p");
					status = STATUS.SPIN_LEFT;
				}
			} else {
				if (!status.equals(STATUS.SPIN_RIGHT)) {
					write("w");
					status = STATUS.SPIN_RIGHT;
				}
			}
		}

		public void putPen() {
			if (pen.equals(PEN.UP)) {
				write("u");
				pen = PEN.DOWN;
			}
		}

		public void endPen() {
			if (pen.equals(PEN.DOWN)) {
				write("v");
				pen = PEN.UP;
			}
		}

		public void stopWheels() {
			if (!status.equals(STATUS.STOP)) {
				write("y");
				status = STATUS.STOP;
			}
		}

		private void write(String cmd) {
			getConnector().write(cmd);
		}
	}

}
